# Copyright (C) 2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable
import textwrap
from typing import Any

from stevedore.example import base


class FieldList(base.FormatterBase):
    """Format values as a reStructuredText field list.

    For example::

      : name1 : value
      : name2 : value
      : name3 : a long value
          will be wrapped with
          a hanging indent
    """

    def format(self, data: dict[str, Any]) -> Iterable[str]:
        """Format the data and return unicode text.

        :param data: A dictionary with string keys and simple types as values.
        :returns: Iterable producing the formatted text.
        """
        for name, value in sorted(data.items()):
            full_text = f': {name} : {value}'
            wrapped_text = textwrap.fill(
                full_text,
                initial_indent='',
                subsequent_indent='    ',
                width=self.max_width,
            )
            yield wrapped_text + '\n'



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Use a cache layer in front of entry point scanning."""

import errno
import glob
import hashlib
import importlib.metadata
import itertools
import json
import logging
import os
import os.path
import struct
import sys
from typing import TypeAlias
from typing import TypedDict

log = logging.getLogger('stevedore._cache')


def _get_cache_dir() -> str:
    """Locate a platform-appropriate cache directory to use.

    Does not ensure that the cache directory exists.
    """
    # Linux, Unix, AIX, etc.
    if os.name == 'posix' and sys.platform != 'darwin':
        # use ~/.cache if empty OR not set
        base_path = os.environ.get('XDG_CACHE_HOME') or os.path.expanduser(
            '~/.cache'
        )
        return os.path.join(base_path, 'python-entrypoints')

    # Mac OS
    elif sys.platform == 'darwin':
        return os.path.expanduser('~/Library/Caches/Python Entry Points')

    # Windows (hopefully)
    else:
        base_path = os.environ.get('LOCALAPPDATA') or os.path.expanduser(
            '~\\AppData\\Local'
        )
        return os.path.join(base_path, 'Python Entry Points')


def _get_mtime(name: str) -> float:
    try:
        s = os.stat(name)
        return s.st_mtime
    except OSError as err:
        if err.errno not in {errno.ENOENT, errno.ENOTDIR}:
            raise
    return -1.0


def _ftobytes(f: float) -> bytes:
    return struct.Struct('f').pack(f)


_PathHashEntryT: TypeAlias = list[tuple[str, float]]
_PathHashSettingsT: TypeAlias = tuple[str, _PathHashEntryT]


def _hash_settings_for_path(path: tuple[str, ...]) -> _PathHashSettingsT:
    """Return a hash and the path settings that created it."""
    paths = []
    h = hashlib.sha256()

    # Tie the cache to the python interpreter, in case it is part of a
    # virtualenv.
    h.update(sys.executable.encode('utf-8'))
    h.update(sys.prefix.encode('utf-8'))

    for entry in path:
        mtime = _get_mtime(entry)
        h.update(entry.encode('utf-8'))
        h.update(_ftobytes(mtime))
        paths.append((entry, mtime))

        for ep_file in itertools.chain(
            glob.iglob(os.path.join(entry, '*.dist-info', 'entry_points.txt')),
            glob.iglob(os.path.join(entry, '*.egg-info', 'entry_points.txt')),
        ):
            mtime = _get_mtime(ep_file)
            h.update(ep_file.encode('utf-8'))
            h.update(_ftobytes(mtime))
            paths.append((ep_file, mtime))

    return (h.hexdigest(), paths)


_CacheEntry = TypedDict(
    '_CacheEntry',
    {
        'groups': dict[str, list[tuple[str, str, str]]],
        'sys.executable': str,
        'sys.prefix': str,
        'path_values': _PathHashEntryT,
    },
)


def _build_cacheable_data() -> _CacheEntry:
    entry_points = importlib.metadata.entry_points()
    groups: dict[str, list[tuple[str, str, str]]] = {}
    for group in entry_points.groups:
        existing = set()
        groups[group] = []
        for ep in entry_points.select(group=group):
            # Filter out duplicates that can occur when testing a
            # package that provides entry points using tox, where the
            # package is installed in the virtualenv that tox builds
            # and is present in the path as '.'.
            item = ep.name, ep.value, ep.group  # convert to tuple
            if item in existing:
                continue
            existing.add(item)
            groups[group].append(item)

    return {
        'groups': groups,
        'sys.executable': sys.executable,
        'sys.prefix': sys.prefix,
        'path_values': [],
    }


class Cache:
    def __init__(self, cache_dir: str | None = None) -> None:
        if cache_dir is None:
            cache_dir = _get_cache_dir()
        self._dir = cache_dir
        self._internal: dict[tuple[str, ...], _CacheEntry] = {}
        self._disable_caching = False

        # Caching can be disabled by either placing .disable file into the
        # target directory or when python executable is under /tmp (this is the
        # case when executed from ansible)
        if any(
            [
                os.path.isfile(os.path.join(self._dir, '.disable')),
                sys.executable[0:4] == '/tmp',  # noqa: S108,
            ]
        ):
            self._disable_caching = True

    def _get_data_for_path(self, path: tuple[str, ...] | None) -> _CacheEntry:
        internal_key = tuple(sys.path) if path is None else tuple(path)

        if internal_key in self._internal:
            return self._internal[internal_key]

        digest, path_values = _hash_settings_for_path(internal_key)
        filename = os.path.join(self._dir, digest)
        try:
            log.debug('reading %s', filename)
            with open(filename) as f:
                data: _CacheEntry = json.load(f)
        except (OSError, json.JSONDecodeError):
            data = _build_cacheable_data()
            data['path_values'] = path_values
            if not self._disable_caching:
                try:
                    log.debug('writing to %s', filename)
                    os.makedirs(self._dir, exist_ok=True)
                    with open(filename, 'w') as f:
                        json.dump(data, f)
                except OSError:
                    # Could not create cache dir or write file.
                    pass

        self._internal[internal_key] = data
        return data

    def get_group_all(
        self, group: str, path: tuple[str, ...] | None = None
    ) -> list[importlib.metadata.EntryPoint]:
        result = []
        data = self._get_data_for_path(path)
        group_data = data.get('groups', {}).get(group, [])
        for vals in group_data:
            result.append(importlib.metadata.EntryPoint(*vals))
        return result

    def get_group_named(
        self, group: str, path: tuple[str, ...] | None = None
    ) -> dict[str, importlib.metadata.EntryPoint]:
        result = {}
        for ep in self.get_group_all(group, path=path):
            if ep.name not in result:
                result[ep.name] = ep
        return result

    def get_single(
        self, group: str, name: str, path: tuple[str, ...] | None = None
    ) -> importlib.metadata.EntryPoint:
        for name, ep in self.get_group_named(group, path=path).items():
            if name == name:
                return ep
        raise ValueError(f'No entrypoint {group!r} in group {name!r}')


_c = Cache()
get_group_all = _c.get_group_all
get_group_named = _c.get_group_named
get_single = _c.get_single



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Sequence
import importlib.metadata
import inspect
from typing import Any
from typing import TypeVar

from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils.statemachine import StringList
from sphinx.application import Sphinx
from sphinx.util import logging
from sphinx.util.nodes import nested_parse_with_titles

from stevedore import extension

LOG = logging.getLogger(__name__)

T = TypeVar('T')


def _get_docstring(plugin: Callable[..., T]) -> str:
    return inspect.getdoc(plugin) or ''


def _simple_list(
    mgr: extension.ExtensionManager[T],
) -> Iterable[tuple[str, str]]:
    for name in sorted(mgr.names()):
        ext = mgr[name]
        doc = _get_docstring(ext.plugin) or '\n'
        summary = doc.splitlines()[0].strip()
        yield (f'* {ext.name} -- {summary}', ext.module_name)


def _detailed_list(
    mgr: extension.ExtensionManager[T],
    over: str = '',
    under: str = '-',
    titlecase: bool = False,
) -> Iterable[tuple[str, str]]:
    for name in sorted(mgr.names()):
        ext = mgr[name]
        if over:
            yield (over * len(ext.name), ext.module_name)
        if titlecase:
            yield (ext.name.title(), ext.module_name)
        else:
            yield (ext.name, ext.module_name)
        if under:
            yield (under * len(ext.name), ext.module_name)
        yield ('\n', ext.module_name)
        doc = _get_docstring(ext.plugin)
        if doc:
            yield (doc, ext.module_name)
        else:
            yield (
                f'.. warning:: No documentation found for {ext.name} in '
                f'{ext.entry_point_target}',
                ext.module_name,
            )
        yield ('\n', ext.module_name)


class ListPluginsDirective(rst.Directive):
    """Present a simple list of the plugins in a namespace."""

    option_spec = {
        'class': directives.class_option,
        'detailed': directives.flag,
        'titlecase': directives.flag,
        'overline-style': directives.single_char_or_unicode,
        'underline-style': directives.single_char_or_unicode,
    }

    has_content = True

    def run(self) -> Sequence[nodes.Node]:
        namespace = ' '.join(self.content).strip()
        LOG.info('documenting plugins from %r', namespace)
        overline_style = self.options.get('overline-style', '')
        underline_style = self.options.get('underline-style', '=')

        def report_load_failure(
            mgr: extension.ExtensionManager[T],
            ep: importlib.metadata.EntryPoint,
            err: BaseException,
        ) -> None:
            LOG.warning('Failed to load %s: %s', ep.module, err)

        mgr: extension.ExtensionManager[Any]
        mgr = extension.ExtensionManager(
            namespace, on_load_failure_callback=report_load_failure
        )

        result = StringList()

        titlecase = 'titlecase' in self.options

        if 'detailed' in self.options:
            data = _detailed_list(
                mgr,
                over=overline_style,
                under=underline_style,
                titlecase=titlecase,
            )
        else:
            data = _simple_list(mgr)
        for text, source in data:
            for line in text.splitlines():
                result.append(line, source)

        # Parse what we have into a new section.
        node = nodes.section()
        node.document = self.state.document
        nested_parse_with_titles(self.state, result, node)

        return node.children


def setup(app: Sphinx) -> dict[str, Any]:
    LOG.info('loading stevedore.sphinxext')
    app.add_directive('list-plugins', ListPluginsDirective)
    return {'parallel_read_safe': True, 'parallel_write_safe': True}



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from typing import Any
from typing import TypeVar

from .extension import ConflictResolverT
from .extension import Extension
from .extension import ignore_conflicts
from .extension import OnLoadFailureCallbackT
from .named import NamedExtensionManager
from .named import OnMissingEntrypointsCallbackT

T = TypeVar('T')


class HookManager(NamedExtensionManager[T]):
    """Coordinate execution of multiple extensions using a common name.

    :param namespace: The namespace for the entry points.
    :param name: The name of the hooks to load.
    :param invoke_on_load: Boolean controlling whether to invoke the
        object returned by the entry point after the driver is loaded.
    :param invoke_args: Positional arguments to pass when invoking
        the object returned by the entry point. Only used if invoke_on_load
        is True.
    :param invoke_kwds: Named arguments to pass when invoking
        the object returned by the entry point. Only used if invoke_on_load
        is True.
    :param on_load_failure_callback: Callback function that will be called when
        an entrypoint can not be loaded. The arguments that will be provided
        when this is called (when an entrypoint fails to load) are
        (manager, entrypoint, exception)
    :param on_missing_entrypoints_callback: Callback function that will be
        called when one or more names cannot be found. The provided argument
        will be a subset of the 'names' parameter.
    :param verify_requirements: **DEPRECATED** This is a no-op and will be
        removed in a future version.
    :param warn_on_missing_entrypoint: **DEPRECATED** Flag to control whether
        failing to load a plugin is reported via a log mess. Only applies if
        on_missing_entrypoints_callback is None. Users should instead set
        ``on_missing_entrypoints_callback`` to ``None`` if they wish to disable
        logging.
    :param conflict_resolver: A callable that determines what to do in the
        event that there are multiple entrypoints in the same group with the
        same name. This is only used if retrieving entrypoint by name.
    """

    def __init__(
        self,
        namespace: str,
        name: str,
        invoke_on_load: bool = False,
        invoke_args: tuple[Any, ...] | None = None,
        invoke_kwds: dict[str, Any] | None = None,
        on_load_failure_callback: 'OnLoadFailureCallbackT[T] | None' = None,
        # NOTE(dhellmann): This default is different from the
        # base class because for hooks it is less likely to
        # be an error to have no entry points present.
        on_missing_entrypoints_callback: (
            OnMissingEntrypointsCallbackT | None
        ) = None,
        verify_requirements: bool | None = None,
        warn_on_missing_entrypoint: bool | None = None,
        *,
        conflict_resolver: 'ConflictResolverT[T]' = ignore_conflicts,
    ):
        invoke_args = () if invoke_args is None else invoke_args
        invoke_kwds = {} if invoke_kwds is None else invoke_kwds

        super().__init__(
            namespace,
            [name],
            invoke_on_load=invoke_on_load,
            invoke_args=invoke_args,
            invoke_kwds=invoke_kwds,
            on_load_failure_callback=on_load_failure_callback,
            on_missing_entrypoints_callback=on_missing_entrypoints_callback,
            verify_requirements=verify_requirements,
            warn_on_missing_entrypoint=warn_on_missing_entrypoint,
        )

    @property
    def _name(self) -> str:
        return self._names[0]

    def __getitem__(  # type: ignore[override]
        self, name: str
    ) -> list[Extension[T]]:
        """Return the named extensions.

        Accessing a HookManager as a dictionary (``em['name']``)
        produces a list of the :class:`Extension` instance(s) with the
        specified name, in the order they would be invoked by map().
        """
        if name != self._name:
            raise KeyError(name)
        return self.extensions



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from collections.abc import Callable
import importlib.metadata
import logging
from typing import Any
from typing import TypeVar

from .extension import ConflictResolverT
from .extension import Extension
from .extension import ExtensionManager
from .extension import ignore_conflicts
from .extension import OnLoadFailureCallbackT

LOG = logging.getLogger(__name__)

T = TypeVar('T')


class EnabledExtensionManager(ExtensionManager[T]):
    """Loads only plugins that pass a check function.

    The check_func argument should return a boolean, with ``True``
    indicating that the extension should be loaded and made available
    and ``False`` indicating that the extension should be ignored.

    :param namespace: The namespace for the entry points.
    :param check_func: Function to determine which extensions to load.
    :param invoke_on_load: Boolean controlling whether to invoke the
        object returned by the entry point after the driver is loaded.
    :param invoke_args: Positional arguments to pass when invoking
        the object returned by the entry point. Only used if invoke_on_load
        is True.
    :param invoke_kwds: Named arguments to pass when invoking
        the object returned by the entry point. Only used if invoke_on_load
        is True.
    :param propagate_map_exceptions: Boolean controlling whether exceptions
        are propagated up through the map call or whether they are logged and
        then ignored
    :param on_load_failure_callback: Callback function that will be called when
        an entrypoint can not be loaded. The arguments that will be provided
        when this is called (when an entrypoint fails to load) are
        (manager, entrypoint, exception)
    :param verify_requirements: **DEPRECATED** This is a no-op and will be
        removed in a future version.
    :param conflict_resolver: A callable that determines what to do in the
        event that there are multiple entrypoints in the same group with the
        same name. This is only used if retrieving entrypoint by name.
    """

    def __init__(
        self,
        namespace: str,
        check_func: Callable[[Extension[T]], bool],
        invoke_on_load: bool = False,
        invoke_args: tuple[Any, ...] | None = None,
        invoke_kwds: dict[str, Any] | None = None,
        propagate_map_exceptions: bool = False,
        on_load_failure_callback: 'OnLoadFailureCallbackT[T] | None' = None,
        verify_requirements: bool | None = None,
        *,
        conflict_resolver: 'ConflictResolverT[T]' = ignore_conflicts,
    ):
        invoke_args = () if invoke_args is None else invoke_args
        invoke_kwds = {} if invoke_kwds is None else invoke_kwds

        self.check_func = check_func

        super().__init__(
            namespace,
            invoke_on_load=invoke_on_load,
            invoke_args=invoke_args,
            invoke_kwds=invoke_kwds,
            propagate_map_exceptions=propagate_map_exceptions,
            on_load_failure_callback=on_load_failure_callback,
            verify_requirements=verify_requirements,
            conflict_resolver=conflict_resolver,
        )

    def _load_one_plugin(
        self,
        ep: importlib.metadata.EntryPoint,
        invoke_on_load: bool,
        invoke_args: tuple[Any, ...],
        invoke_kwds: dict[str, Any],
    ) -> Extension[T] | None:
        ext = super()._load_one_plugin(
            ep, invoke_on_load, invoke_args, invoke_kwds
        )
        if ext and not self.check_func(ext):
            LOG.debug('ignoring extension %r', ep.name)
            return None
        return ext



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Sequence
import importlib.metadata
import logging
from typing import Any
from typing import TYPE_CHECKING
from typing import TypeVar
import warnings

from .extension import ConflictResolverT
from .extension import Extension
from .extension import ExtensionManager
from .extension import ignore_conflicts
from .extension import OnLoadFailureCallbackT

if TYPE_CHECKING:
    from typing_extensions import Self

LOG = logging.getLogger(__name__)

T = TypeVar('T')
OnMissingEntrypointsCallbackT = Callable[[Iterable[str]], None]


def warning_on_missing_entrypoint(missing_names: Iterable[str]) -> None:
    LOG.warning('Could not load %s', ', '.join(missing_names))


class NamedExtensionManager(ExtensionManager[T]):
    """Loads only the named extensions.

    This is useful for explicitly enabling extensions in a
    configuration file, for example.

    :param namespace: The namespace for the entry points.
    :param names: The names of the extensions to load.
    :param invoke_on_load: Boolean controlling whether to invoke the
        object returned by the entry point after the driver is loaded.
    :param invoke_args: Positional arguments to pass when invoking
        the object returned by the entry point. Only used if invoke_on_load
        is True.
    :param invoke_kwds: Named arguments to pass when invoking
        the object returned by the entry point. Only used if invoke_on_load
        is True.
    :param name_order: If true, sort the loaded extensions to match the
        order used in ``names``.
    :param propagate_map_exceptions: Boolean controlling whether exceptions
        are propagated up through the map call or whether they are logged and
        then ignored
    :param on_load_failure_callback: Callback function that will be called when
        an entrypoint can not be loaded. The arguments that will be provided
        when this is called (when an entrypoint fails to load) are
        (manager, entrypoint, exception)
    :param on_missing_entrypoints_callback: Callback function that will be
        called when one or more names cannot be found. The provided argument
        will be a subset of the 'names' parameter.
    :param verify_requirements: **DEPRECATED** This is a no-op and will be
        removed in a future version.
    :param warn_on_missing_entrypoint: **DEPRECATED** Flag to control whether
        failing to load a plugin is reported via a log mess. Only applies if
        on_missing_entrypoints_callback is None. Users should instead set
        ``on_missing_entrypoints_callback`` to ``None`` if they wish to disable
        logging.
    :param conflict_resolver: A callable that determines what to do in the
        event that there are multiple entrypoints in the same group with the
        same name. This is only used if retrieving entrypoint by name.
    """

    def __init__(
        self,
        namespace: str,
        names: Sequence[str],
        invoke_on_load: bool = False,
        invoke_args: tuple[Any, ...] | None = None,
        invoke_kwds: dict[str, Any] | None = None,
        name_order: bool = False,
        propagate_map_exceptions: bool = False,
        on_load_failure_callback: 'OnLoadFailureCallbackT[T] | None' = None,
        on_missing_entrypoints_callback: (
            OnMissingEntrypointsCallbackT | None
        ) = warning_on_missing_entrypoint,
        verify_requirements: bool | None = None,
        warn_on_missing_entrypoint: bool | None = None,
        *,
        conflict_resolver: 'ConflictResolverT[T]' = ignore_conflicts,
    ) -> None:
        self._names = names
        self._missing_names: set[str] = set()
        self._name_order = name_order

        if warn_on_missing_entrypoint is not None:
            warnings.warn(
                "The warn_on_missing_entrypoint option is deprecated for "
                "removal. If you wish to disable warnings, you should instead "
                "override 'on_missing_entrypoints_callback'",
                DeprecationWarning,
            )
            if not warn_on_missing_entrypoint:
                on_missing_entrypoints_callback = None

        self._on_missing_entrypoints_callback = on_missing_entrypoints_callback

        super().__init__(
            namespace,
            invoke_on_load=invoke_on_load,
            invoke_args=invoke_args,
            invoke_kwds=invoke_kwds,
            propagate_map_exceptions=propagate_map_exceptions,
            on_load_failure_callback=on_load_failure_callback,
            verify_requirements=verify_requirements,
            conflict_resolver=conflict_resolver,
        )

    @classmethod
    def make_test_instance(
        cls,
        extensions: list[Extension[T]],
        namespace: str = 'TESTING',
        propagate_map_exceptions: bool = False,
        on_load_failure_callback: 'OnLoadFailureCallbackT[T] | None' = None,
        verify_requirements: bool | None = None,
        *,
        conflict_resolver: 'ConflictResolverT[T]' = ignore_conflicts,
    ) -> 'Self':
        """Construct a test NamedExtensionManager

        Test instances are passed a list of extensions to use rather than
        loading them from entry points.

        :param extensions: Pre-configured Extension instances
        :param namespace: The namespace for the manager; used only for
            identification since the extensions are passed in.
        :param propagate_map_exceptions: Boolean controlling whether exceptions
            are propagated up through the map call or whether they are logged
            and then ignored
        :param on_load_failure_callback: Callback function that will
            be called when an entrypoint can not be loaded. The
            arguments that will be provided when this is called (when
            an entrypoint fails to load) are (manager, entrypoint,
            exception)
        :param verify_requirements: **DEPRECATED** This is a no-op and will be
            removed in a future version.
        :param conflict_resolver: A callable that determines what to do in the
            event that there are multiple entrypoints in the same group with
            the same name. This is only used if retrieving entrypoint by name.
        :return: The manager instance, initialized for testing
        """
        if verify_requirements is not None:
            warnings.warn(
                'The verify_requirements argument is now a no-op and is '
                'deprecated for removal. Remove the argument from calls.',
                DeprecationWarning,
            )

        o = cls.__new__(cls)
        o.namespace = namespace
        o.propagate_map_exceptions = propagate_map_exceptions
        o._on_load_failure_callback = on_load_failure_callback
        o._names = [e.name for e in extensions]
        o._missing_names = set()
        o._name_order = False
        o._conflict_resolver = conflict_resolver
        o._init_plugins(extensions)
        return o

    def _init_plugins(self, extensions: list[Extension[T]]) -> None:
        super()._init_plugins(extensions)

        if self._name_order:
            self.extensions = [
                self[n] for n in self._names if n not in self._missing_names
            ]

    def _load_plugins(
        self,
        invoke_on_load: bool,
        invoke_args: tuple[Any, ...],
        invoke_kwds: dict[str, Any],
    ) -> list[Extension[T]]:
        extensions = super()._load_plugins(
            invoke_on_load, invoke_args, invoke_kwds
        )

        self._missing_names = set(self._names) - {e.name for e in extensions}
        if self._missing_names and self._on_missing_entrypoints_callback:
            self._on_missing_entrypoints_callback(self._missing_names)

        return extensions

    def _load_one_plugin(
        self,
        ep: importlib.metadata.EntryPoint,
        invoke_on_load: bool,
        invoke_args: tuple[Any, ...],
        invoke_kwds: dict[str, Any],
    ) -> Extension[T] | None:
        # Check the name before going any further to prevent
        # undesirable code from being loaded at all if we are not
        # going to use it.
        if ep.name not in self._names:
            return None

        return super()._load_one_plugin(
            ep, invoke_on_load, invoke_args, invoke_kwds
        )



# Copyright (C) 2020 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from collections.abc import Iterable
from typing import Any

from stevedore.example import base


class Simple(base.FormatterBase):
    """A very basic formatter."""

    def format(self, data: dict[str, Any]) -> Iterable[str]:
        """Format the data and return unicode text.

        :param data: A dictionary with string keys and simple types as values.
        :returns: Iterable producing the formatted text.
        """
        for name, value in sorted(data.items()):
            line = f'{name} = {value}\n'
            yield line



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""ExtensionManager"""

from collections.abc import Callable
from collections.abc import ItemsView
from collections.abc import Iterator
import importlib.metadata
import itertools
import logging
import operator
from typing import Any
from typing import Concatenate
from typing import Generic
from typing import ParamSpec
from typing import TYPE_CHECKING
from typing import TypeAlias
from typing import TypeVar
import warnings

from . import _cache
from .exception import MultipleMatches
from .exception import NoMatches

if TYPE_CHECKING:
    from typing_extensions import Self

LOG = logging.getLogger(__name__)

T = TypeVar('T')
U = TypeVar('U')
P = ParamSpec('P')


class Extension(Generic[T]):
    """Book-keeping object for tracking extensions.

    The arguments passed to the constructor are saved as attributes of
    the instance using the same names, and can be accessed by the
    callables passed to :meth:`map` or when iterating over an
    :class:`ExtensionManager` directly.

    :param name: The entry point name.
    :param entry_point: The EntryPoint instance returned by :mod:`entrypoints`.
    :param plugin: The value returned by entry_point.load()
    :param obj: The object returned by ``plugin(*args, **kwds)`` if the
                manager invoked the extension on load.

    """

    def __init__(
        self,
        name: str,
        entry_point: importlib.metadata.EntryPoint,
        plugin: Callable[..., T],
        obj: T | None,
    ) -> None:
        self.name = name
        self.entry_point = entry_point
        self.plugin = plugin
        self.obj = obj

    @property
    def module_name(self) -> str:
        """The name of the module from which the entry point is loaded.

        :return: A string in 'dotted.module' format.
        """
        return self.entry_point.module

    @property
    def attr(self) -> str:
        """The attribute of the module to be loaded."""
        return self.entry_point.attr

    @property
    def entry_point_target(self) -> str:
        """The module and attribute referenced by this extension's entry_point.

        :return: A string representation of the target of the entry point in
            'dotted.module:object' format.
        """
        return self.entry_point.value


#: OnLoadFailureCallbackT defines the type for callbacks when a plugin fails
#: to load. The callback callable should expect the extension manager instance,
#: the underlying entrypoint instance, and the exception raised during
#: attempted loading.
OnLoadFailureCallbackT: TypeAlias = Callable[
    ['ExtensionManager[T]', importlib.metadata.EntryPoint, BaseException], None
]

#: ConflictResolver defines the type for conflict resolution callables. The
#: callable should expect the extension namespace, extension name, and a list
#: of the entrypoints themselves.
ConflictResolverT: TypeAlias = Callable[
    [str, str, list[Extension[T]]], Extension[T]
]


def ignore_conflicts(
    namespace: str, name: str, entrypoints: list[Extension[T]]
) -> Extension[T]:
    LOG.warning(
        "multiple implementations found for the '%(name)s' extension in "
        "%(namespace)s namespace: %(conflicts)s",
        {
            'name': name,
            'namespace': namespace,
            'conflicts': ', '.join(
                ep.plugin.__qualname__ for ep in entrypoints
            ),
        },
    )
    # use the most last found entrypoint
    return entrypoints[-1]


def error_on_conflict(
    namespace: str, name: str, entrypoints: list[Extension[T]]
) -> Extension[T]:
    raise MultipleMatches(
        "multiple implementations found for the '{name}' command in "
        "{namespace} namespace: {conflicts}".format(
            name=name,
            namespace=namespace,
            conflicts=', '.join(ep.plugin.__qualname__ for ep in entrypoints),
        )
    )


class ExtensionManager(Generic[T]):
    """Base class for all of the other managers.

    :param namespace: The namespace for the entry points.
    :param invoke_on_load: Boolean controlling whether to invoke the
        object returned by the entry point after the driver is loaded.
    :param invoke_args: Positional arguments to pass when invoking
        the object returned by the entry point. Only used if invoke_on_load
        is True.
    :param invoke_kwds: Named arguments to pass when invoking
        the object returned by the entry point. Only used if invoke_on_load
        is True.
    :param propagate_map_exceptions: Boolean controlling whether exceptions
        are propagated up through the map call or whether they are logged and
        then ignored
    :param on_load_failure_callback: Callback function that will be called when
        an entrypoint can not be loaded. The arguments that will be provided
        when this is called (when an entrypoint fails to load) are
        (manager, entrypoint, exception)
    :param verify_requirements: **DEPRECATED** This is a no-op and will be
        removed in a future version.
    :param conflict_resolver: A callable that determines what to do in the
        event that there are multiple entrypoints in the same group with the
        same name. This is only used if retrieving entrypoint by name.
    """

    ENTRY_POINT_CACHE: dict[str, list[importlib.metadata.EntryPoint]] = {}

    def __init__(
        self,
        namespace: str,
        invoke_on_load: bool = False,
        invoke_args: tuple[Any, ...] | None = None,
        invoke_kwds: dict[str, Any] | None = None,
        propagate_map_exceptions: bool = False,
        on_load_failure_callback: 'OnLoadFailureCallbackT[T] | None' = None,
        verify_requirements: bool | None = None,
        *,
        conflict_resolver: 'ConflictResolverT[T]' = ignore_conflicts,
    ) -> None:
        invoke_args = () if invoke_args is None else invoke_args
        invoke_kwds = {} if invoke_kwds is None else invoke_kwds

        if verify_requirements is not None:
            warnings.warn(
                'The verify_requirements argument is now a no-op and is '
                'deprecated for removal. Remove the argument from calls.',
                DeprecationWarning,
            )

        self.namespace = namespace
        self.propagate_map_exceptions = propagate_map_exceptions
        self._on_load_failure_callback = on_load_failure_callback
        self._conflict_resolver = conflict_resolver

        extensions = self._load_plugins(
            invoke_on_load, invoke_args, invoke_kwds
        )

        self._init_plugins(extensions)

    @classmethod
    def make_test_instance(
        cls,
        extensions: list[Extension[T]],
        namespace: str = 'TESTING',
        propagate_map_exceptions: bool = False,
        on_load_failure_callback: 'OnLoadFailureCallbackT[T] | None' = None,
        verify_requirements: bool | None = None,
        *,
        conflict_resolver: 'ConflictResolverT[T]' = ignore_conflicts,
    ) -> 'Self':
        """Construct a test ExtensionManager

        Test instances are passed a list of extensions to work from rather
        than loading them from entry points.

        :param extensions: Pre-configured Extension instances to use
        :param namespace: The namespace for the manager; used only for
            identification since the extensions are passed in.
        :param propagate_map_exceptions: When calling map, controls whether
            exceptions are propagated up through the map call or whether they
            are logged and then ignored
        :param on_load_failure_callback: Callback function that will
            be called when an entrypoint can not be loaded. The
            arguments that will be provided when this is called (when
            an entrypoint fails to load) are (manager, entrypoint,
            exception)
        :param verify_requirements: **DEPRECATED** This is a no-op and will be
            removed in a future version.
        :param conflict_resolver: A callable that determines what to do in the
            event that there are multiple entrypoints in the same group with
            the same name. This is only used if retrieving entrypoint by name.
        :return: The manager instance, initialized for testing
        """
        if verify_requirements is not None:
            warnings.warn(
                'The verify_requirements argument is now a no-op and is '
                'deprecated for removal. Remove the argument from calls.',
                DeprecationWarning,
            )

        o = cls.__new__(cls)
        o.namespace = namespace
        o.propagate_map_exceptions = propagate_map_exceptions
        o._on_load_failure_callback = on_load_failure_callback
        o._conflict_resolver = conflict_resolver
        o._init_plugins(extensions)
        return o

    def _init_plugins(self, extensions: list[Extension[T]]) -> None:
        self.extensions: list[Extension[T]] = extensions
        self._extensions_by_name_cache: dict[str, Extension[T]] | None = None

    @property
    def _extensions_by_name(self) -> dict[str, Extension[T]]:
        if self._extensions_by_name_cache is None:
            d = {}
            for name, _extensions in itertools.groupby(
                self.extensions, lambda x: x.name
            ):
                extensions = list(_extensions)
                if len(extensions) > 1:
                    ext = self._conflict_resolver(
                        self.namespace, name, extensions
                    )
                else:
                    ext = extensions[0]

                d[name] = ext

            self._extensions_by_name_cache = d
        return self._extensions_by_name_cache

    def list_entry_points(self) -> list[importlib.metadata.EntryPoint]:
        """Return the list of entry points for this namespace.

        The entry points are not actually loaded, their list is just read and
        returned.
        """
        if self.namespace not in self.ENTRY_POINT_CACHE:
            eps = list(_cache.get_group_all(self.namespace))
            self.ENTRY_POINT_CACHE[self.namespace] = eps
        return self.ENTRY_POINT_CACHE[self.namespace]

    def entry_points_names(self) -> list[str]:
        """Return the list of entry points names for this namespace."""
        return list(map(operator.attrgetter("name"), self.list_entry_points()))

    def _load_plugins(
        self,
        invoke_on_load: bool,
        invoke_args: tuple[Any, ...],
        invoke_kwds: dict[str, Any],
    ) -> list[Extension[T]]:
        extensions = []
        for ep in self.list_entry_points():
            LOG.debug('found extension %r', ep)
            try:
                ext = self._load_one_plugin(
                    ep, invoke_on_load, invoke_args, invoke_kwds
                )
                if ext:
                    extensions.append(ext)
            except (KeyboardInterrupt, AssertionError):
                raise
            except Exception as err:
                if self._on_load_failure_callback is not None:
                    self._on_load_failure_callback(self, ep, err)
                else:
                    # Log the reason we couldn't import the module,
                    # usually without a traceback. The most common
                    # reason is an ImportError due to a missing
                    # dependency, and the error message should be
                    # enough to debug that.  If debug logging is
                    # enabled for our logger, provide the full
                    # traceback.
                    LOG.error(
                        'Could not load %r: %s',
                        ep.name,
                        err,
                        exc_info=LOG.isEnabledFor(logging.DEBUG),
                    )
        return extensions

    # NOTE(stephenfin): While this can't return None, all the subclasses can,
    # and this allows us to satisfy Liskov's Principle. `_load_plugins` handles
    # things just fine in either case.
    def _load_one_plugin(
        self,
        ep: importlib.metadata.EntryPoint,
        invoke_on_load: bool,
        invoke_args: tuple[Any],
        invoke_kwds: dict[str, Any],
    ) -> Extension[T] | None:
        plugin = ep.load()
        if invoke_on_load:
            obj = plugin(*invoke_args, **invoke_kwds)
        else:
            obj = None
        return Extension(ep.name, ep, plugin, obj)

    def names(self) -> list[str]:
        """Returns the names of the discovered extensions"""
        # We want to return the names of the extensions in the order
        # they would be used by map(), since some subclasses change
        # that order.
        return [e.name for e in self.extensions]

    def map(
        self,
        func: Callable[Concatenate[Extension[T], P], U],
        *args: P.args,
        **kwds: P.kwargs,
    ) -> list[U]:
        """Iterate over the extensions invoking func() for each.

        The signature for func() should be::

            def func(ext, *args, **kwds):
                pass

        The first argument to func(), 'ext', is the
        :class:`~stevedore.extension.Extension` instance.

        Exceptions raised from within func() are propagated up and
        processing stopped if self.propagate_map_exceptions is True,
        otherwise they are logged and ignored.

        :param func: Callable to invoke for each extension.
        :param args: Variable arguments to pass to func()
        :param kwds: Keyword arguments to pass to func()
        :returns: List of values returned from func()
        """
        if not self.extensions:
            # FIXME: Use a more specific exception class here.
            raise NoMatches(f'No {self.namespace} extensions found')
        response: list[U] = []
        for e in self.extensions:
            self._invoke_one_plugin(response.append, func, e, *args, **kwds)
        return response

    @staticmethod
    def _call_extension_method(
        extension: Extension[T], /, method_name: str, *args: Any, **kwds: Any
    ) -> Any:
        return getattr(extension.obj, method_name)(*args, **kwds)

    def map_method(self, method_name: str, *args: Any, **kwds: Any) -> Any:
        """Iterate over the extensions invoking a method by name.

        This is equivalent of using :meth:`map` with func set to
        `lambda x: x.obj.method_name()`
        while being more convenient.

        Exceptions raised from within the called method are propagated up
        and processing stopped if self.propagate_map_exceptions is True,
        otherwise they are logged and ignored.

        .. versionadded:: 0.12

        :param method_name: The extension method name
                            to call for each extension.
        :param args: Variable arguments to pass to method
        :param kwds: Keyword arguments to pass to method
        :returns: List of values returned from methods
        """
        return self.map(
            self._call_extension_method, method_name, *args, **kwds
        )

    def _invoke_one_plugin(
        self,
        response_callback: Callable[..., Any],
        func: Callable[Concatenate[Extension[T], P], U],
        e: Extension[T],
        *args: P.args,
        **kwds: P.kwargs,
    ) -> None:
        try:
            response_callback(func(e, *args, **kwds))
        except Exception as err:
            if self.propagate_map_exceptions:
                raise
            else:
                LOG.error('error calling %r: %s', e.name, err)
                LOG.exception(err)

    def items(self) -> ItemsView[str, Extension[T]]:
        """Return an iterator of tuples of the form (name, extension).

        This is analogous to the Mapping.items() method.
        """
        return self._extensions_by_name.items()

    def __iter__(self) -> Iterator[Extension[T]]:
        """Produce iterator for the manager.

        Iterating over an ExtensionManager produces the :class:`Extension`
        instances in the order they would be invoked.
        """
        return iter(self.extensions)

    def __getitem__(self, name: str) -> Extension[T]:
        """Return the named extension.

        Accessing an ExtensionManager as a dictionary (``em['name']``)
        produces the :class:`Extension` instance with the specified name.
        """
        return self._extensions_by_name[name]

    def __contains__(self, name: str) -> bool:
        """Return true if name is in list of enabled extensions."""
        return any(extension.name == name for extension in self.extensions)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from collections.abc import Callable
import importlib.metadata
from typing import Any
from typing import Concatenate
from typing import ParamSpec
from typing import TYPE_CHECKING
from typing import TypeVar

from .exception import MultipleMatches
from .exception import NoMatches
from .extension import ConflictResolverT
from .extension import Extension
from .extension import ExtensionManager
from .extension import ignore_conflicts
from .extension import OnLoadFailureCallbackT
from .named import NamedExtensionManager
from .named import OnMissingEntrypointsCallbackT
from .named import warning_on_missing_entrypoint

if TYPE_CHECKING:
    from typing_extensions import Self

T = TypeVar('T')
U = TypeVar('U')
P = ParamSpec('P')


class DriverManager(NamedExtensionManager[T]):
    """Load a single plugin with a given name from the namespace.

    :param namespace: The namespace for the entry points.
    :param name: The name of the driver to load.
    :param invoke_on_load: Boolean controlling whether to invoke the
        object returned by the entry point after the driver is loaded.
    :param invoke_args: Positional arguments to pass when invoking
        the object returned by the entry point. Only used if invoke_on_load
        is True.
    :param invoke_kwds: Named arguments to pass when invoking
        the object returned by the entry point. Only used if invoke_on_load
        is True.
    :param on_load_failure_callback: Callback function that will be called when
        an entrypoint can not be loaded. The arguments that will be provided
        when this is called (when an entrypoint fails to load) are
        (manager, entrypoint, exception)
    :param on_missing_entrypoints_callback: Callback function that will be
        called when one or more names cannot be found. The provided argument
        will be a subset of the 'names' parameter.
    :param verify_requirements: **DEPRECATED** This is a no-op and will be
        removed in a future version.
    :param warn_on_missing_entrypoint: **DEPRECATED** Flag to control whether
        failing to load a plugin is reported via a log mess. Only applies if
        on_missing_entrypoints_callback is None. Users should instead set
        ``on_missing_entrypoints_callback`` to ``None`` if they wish to disable
        logging.
    :param conflict_resolver: A callable that determines what to do in the
        event that there are multiple entrypoints in the same group with the
        same name. This is only used if retrieving entrypoint by name.
    """

    def __init__(
        self,
        namespace: str,
        name: str,
        invoke_on_load: bool = False,
        invoke_args: tuple[Any, ...] | None = None,
        invoke_kwds: dict[str, Any] | None = None,
        on_load_failure_callback: 'OnLoadFailureCallbackT[T] | None' = None,
        on_missing_entrypoints_callback: (
            OnMissingEntrypointsCallbackT | None
        ) = warning_on_missing_entrypoint,
        verify_requirements: bool | None = None,
        warn_on_missing_entrypoint: bool | None = None,
        *,
        conflict_resolver: 'ConflictResolverT[T]' = ignore_conflicts,
    ) -> None:
        invoke_args = () if invoke_args is None else invoke_args
        invoke_kwds = {} if invoke_kwds is None else invoke_kwds
        on_load_failure_callback = (
            on_load_failure_callback or self._default_on_load_failure
        )

        super().__init__(
            namespace=namespace,
            names=[name],
            invoke_on_load=invoke_on_load,
            invoke_args=invoke_args,
            invoke_kwds=invoke_kwds,
            on_load_failure_callback=on_load_failure_callback,
            verify_requirements=verify_requirements,
            warn_on_missing_entrypoint=warn_on_missing_entrypoint,
            conflict_resolver=conflict_resolver,
        )

    @staticmethod
    def _default_on_load_failure(
        manager: 'ExtensionManager[T]',
        ep: importlib.metadata.EntryPoint,
        err: BaseException,
    ) -> None:
        raise

    @classmethod
    def make_test_instance(  # type: ignore[override]
        cls,
        extension: Extension[T],
        namespace: str = 'TESTING',
        propagate_map_exceptions: bool = False,
        on_load_failure_callback: 'OnLoadFailureCallbackT[T] | None' = None,
        verify_requirements: bool | None = None,
        *,
        conflict_resolver: 'ConflictResolverT[T]' = ignore_conflicts,
    ) -> 'Self':
        """Construct a test DriverManager

        Test instances are passed a list of extensions to work from rather
        than loading them from entry points.

        :param extension: Pre-configured Extension instance
        :param namespace: The namespace for the manager; used only for
            identification since the extensions are passed in.
        :param propagate_map_exceptions: Boolean controlling whether exceptions
            are propagated up through the map call or whether they are logged
            and then ignored
        :param on_load_failure_callback: Callback function that will
            be called when an entrypoint can not be loaded. The
            arguments that will be provided when this is called (when
            an entrypoint fails to load) are (manager, entrypoint,
            exception)
        :param verify_requirements: **DEPRECATED** This is a no-op and will be
            removed in a future version.
        :param conflict_resolver: A callable that determines what to do in the
            event that there are multiple entrypoints in the same group with
            the same name. This is only used if retrieving entrypoint by name.
        :return: The manager instance, initialized for testing
        """
        o = super().make_test_instance(
            [extension],
            namespace=namespace,
            propagate_map_exceptions=propagate_map_exceptions,
            on_load_failure_callback=on_load_failure_callback,
            verify_requirements=verify_requirements,
            conflict_resolver=conflict_resolver,
        )
        return o

    def _init_plugins(self, extensions: list[Extension[T]]) -> None:
        super()._init_plugins(extensions)

        if not self.extensions:
            name = self._names[0]
            raise NoMatches(
                f'No {self.namespace!r} driver found, looking for {name!r}'
            )
        if len(self.extensions) > 1:
            discovered_drivers = ','.join(
                e.entry_point_target for e in self.extensions
            )

            raise MultipleMatches(
                f'Multiple {self.namespace!r} drivers found: '
                f'{discovered_drivers}'
            )

    def __call__(
        self,
        func: Callable[Concatenate[Extension[T], P], U],
        *args: Any,
        **kwds: Any,
    ) -> U | None:
        """Invokes func() for the single loaded extension.

        The signature for func() should be::

            def func(ext, *args, **kwds):
                pass

        The first argument to func(), 'ext', is the
        :class:`~stevedore.extension.Extension` instance.

        Exceptions raised from within func() are logged and ignored.

        :param func: Callable to invoke for each extension.
        :param args: Variable arguments to pass to func()
        :param kwds: Keyword arguments to pass to func()
        :returns: List of values returned from func()
        """
        results = self.map(func, *args, **kwds)
        if results:
            return results[0]
        return None

    @property
    def driver(self) -> T | Callable[..., T]:
        """Returns the driver being used by this manager."""
        ext = self.extensions[0]
        return ext.obj if ext.obj else ext.plugin



# Copyright (C) 2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from collections.abc import Iterable
from typing import Any


class FormatterBase(metaclass=abc.ABCMeta):
    """Base class for example plugin used in the tutorial."""

    def __init__(self, max_width: int = 60) -> None:
        self.max_width = max_width

    @abc.abstractmethod
    def format(self, data: dict[str, Any]) -> Iterable[str]:
        """Format the data and return unicode text.

        :param data: A dictionary with string keys and simple types as values.
        :returns: Iterable producing the formatted text.
        """



# Copyright (C) 2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from collections.abc import Iterable
from typing import Any

from stevedore import extension

from .base import FormatterBase


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--width', default=60, type=int, help='maximum output width for text'
    )
    parsed_args = parser.parse_args()

    data = {'a': 'A', 'b': 'B', 'long': 'word ' * 80}

    mgr: extension.ExtensionManager[FormatterBase]
    mgr = extension.ExtensionManager(
        namespace='stevedore.example.formatter',
        invoke_on_load=True,
        invoke_args=(parsed_args.width,),
    )

    def format_data(
        ext: extension.Extension[FormatterBase], data: dict[str, Any], /
    ) -> tuple[str, Iterable[str]]:
        assert ext.obj is not None
        return (ext.name, ext.obj.format(data))

    results = mgr.map(format_data, data)

    for name, result in results:
        print(f'Formatter: {name}')
        for chunk in result:
            print(chunk, end='')
        print('')
